#!/usr/bin/env python
"""util.py

Utility functions for nanslice module
"""
from pathlib import Path
import numpy as np
from nibabel import load


def check_path(maybe_path):
    """Helper function to check if an object is path-like"""
    if isinstance(maybe_path, Path) or isinstance(maybe_path, str):
        return True
    else:
        return False


def ensure_image(maybe_path):
    """Helper function that lets either images or paths be passed around"""
    if check_path(maybe_path):
        return load(str(maybe_path))
    else:
        return maybe_path


def center_of_mass(img):
    """Calculates the center of mass of the image"""
    idx0 = np.argmax(np.sum(img.get_data(), axis=(1, 2)))
    idx1 = np.argmax(np.sum(img.get_data(), axis=(0, 2)))
    idx2 = np.argmax(np.sum(img.get_data(), axis=(0, 1)))
    phys = np.dot(img.affine, np.array([idx0, idx1, idx2, 1]).T)
    return phys


def add_common_arguments(parser):
    """Defines a set of common arguments that are shared between nanviewer and nanslicer"""
    parser.add_argument('base_image', help='Base (structural image)', type=str)
    parser.add_argument('--mask', type=str,
                        help='Mask image')
    parser.add_argument('--crop_center', type=float, nargs=3,
                        help='Center of crop box (X Y Z)')
    parser.add_argument('--crop_size', type=float, nargs=3,
                        help='Size of crop box (X Y Z)')
    parser.add_argument('--base_map', type=str, default='gist_gray',
                        help='Base image colormap to use from Matplotlib, default = gist_gray')
    parser.add_argument('--base_lims', type=float, nargs=2, default=None,
                        help='Specify base image window')
    parser.add_argument('--base_lims_p', type=float, nargs=2, default=None,
                        help='Specify base image window in percent')
    parser.add_argument('--base_scale', type=float, default=1.0,
                        help='Scaling for base image before mapping, default=1.0')
    parser.add_argument('--base_label', type=str, default='',
                        help='Label for base color axis')

    parser.add_argument('--overlay', type=str,
                        help='Add color overlay')
    parser.add_argument('--overlay_map', type=str, default='RdYlBu_r',
                        help='Overlay colormap, default = RdYlBu_r')
    parser.add_argument('--overlay_lim', type=float, nargs=2, default=(-1, 1),
                        help='Overlay window, default=-1 1')
    parser.add_argument('--overlay_mask', type=str,
                        help='Mask color image')
    parser.add_argument('--overlay_mask_thresh', type=float,
                        help='Overlay mask threshold')
    parser.add_argument('--overlay_scale', type=float, default=1.0,
                        help='Scaling for overlay image before mapping, default=1.0')
    parser.add_argument('--overlay_label', type=str, default='',
                        help='Label for overlay color axis')
    parser.add_argument('--overlay_alpha', type=str,
                        help='Image for transparency-coding of overlay')
    parser.add_argument('--overlay_alpha_lim', type=float, nargs=2, default=(0.5, 1.0),
                        help='Overlay Alpha/transparency window, default=0.5 1.0')
    parser.add_argument('--overlay_alpha_scale', type=float, default=1.0,
                        help='Scaling factor for the alpha image')
    parser.add_argument('--overlay_alpha_label', type=str, default='1-p',
                        help='Label for overlay alpha/transparency axis')
    parser.add_argument('--contour', type=float, action='append',
                        help='Add alpha image contours (can be multiple)')
    parser.add_argument('--contour_color', type=str, action='append', default='k',
                        help='Choose contour colors')
    parser.add_argument('--contour_style', type=str, action='append', default='-',
                        help='Choose contor line-styles')

    parser.add_argument('--samples', type=int, default=128,
                        help='Number of samples for slicing, default=128')
    parser.add_argument('--interp', type=str, default='hanning',
                        help='Display interpolation mode, default=hanning')
    parser.add_argument('--interp_order', type=int, default=1,
                        help='Data interpolation order, default=1')
    parser.add_argument('--orient', type=str, default='clin',
                        help='Clinical (clin) or Pre-clinical (preclin) orientation')
    return parser



#!/usr/bin/env python
"""colorbar.py

Functions to create a colorbar and a dual-axis color/alphabar.

Matplotlib has no concept of an "alphabar". In addition, because the standard
matplotlib colormaps and colorbars only work on scalar (single-channel) input
images, and matplotlib does not deal with alpha/transparency correctly, nanslice
images are true-color RGB arrays. Hence we need to roll our own colorbar as well
"""
import numpy as np
from . import slice_func


def colorbar(axes, cm_name, clims, clabel,
             black_backg=True, show_ticks=True, tick_fmt='{:.4g}', orient='h'):
    """
    Plots a colorbar in the specified axes

    Parameters:

    - axes -- matplotlib axes instance to use for plotting
    - cm_name -- Colormap name
    - clims -- The limits for the colormap & bar
    - clabel -- Label to place on the color axis
    - black_bg -- Boolean indicating if the background to this plot is black, and hence white text/borders should be used
    - show_ticks -- Set to false if you don't want ticks on the color axis
    - tick_fmt -- Valid format string for the tick labels
    - orient -- 'v' or 'h' for whether you want a vertical or horizontal colorbar
    """
    steps = 32
    if orient == 'h':
        ext = (clims[0], clims[1], 0, 1)
        cdata = np.tile(np.linspace(clims[0], clims[1], steps)[
                        np.newaxis, :], [steps, 1])
    else:
        ext = (0, 1, clims[0], clims[1])
        cdata = np.tile(np.linspace(clims[0], clims[1], steps)[
                        :, np.newaxis], [1, steps])
    color = slice_func.colorize(cdata, cm_name, clims)
    axes.imshow(color, origin='lower', interpolation='hanning',
                extent=ext, aspect='auto')
    if black_backg:
        forecolor = 'w'
    else:
        forecolor = 'k'
    if show_ticks:
        ticks = (clims[0], np.sum(clims)/2, clims[1])
        labels = (tick_fmt.format(clims[0]),
                  clabel, tick_fmt.format(clims[1]))
        if orient == 'h':
            axes.set_xticks(ticks)
            axes.set_xticklabels(labels, color=forecolor)
            axes.set_yticks(())
        else:
            axes.set_yticks(ticks)
            axes.set_yticklabels(labels, color=forecolor,
                                 rotation='vertical', va='center')
            axes.yaxis.tick_right()
            axes.set_xticks(())
    else:
        if orient == 'h':
            axes.set_xticks((np.sum(clims)/2,))
            axes.set_xticklabels((clabel,), color=forecolor)
            axes.set_yticks(())
        else:
            axes.set_yticks((np.sum(clims)/2,))
            axes.set_yticklabels((clabel,), color=forecolor)
            axes.yaxis.tick_right()
            axes.set_xticks(())
    axes.tick_params(axis='both', which='both', length=0)
    axes.spines['top'].set_visible(False)
    axes.spines['bottom'].set_visible(False)
    axes.spines['left'].set_visible(False)
    axes.spines['right'].set_visible(False)
    axes.yaxis.label.set_color(forecolor)
    axes.xaxis.label.set_color(forecolor)
    axes.axis('on')


def alphabar(axes, cm_name, clims, clabel,
             alims, alabel, alines=None, alines_colors=('k',), alines_styles=('solid',),
             cfmt='{:.3g}', afmt='{:.3g}',
             black_backg=True, orient='h'):
    """
    Plots a 2D 'alphabar' with color and transparency axes in the specified matplotlib axes object

    Parameters:

    - axes -- matplotlib axes instance to use for plotting
    - cm_name -- Colormap name
    - clims -- The limits for the colormap & bar
    - clabel -- Label to place on the color axis
    - alims -- The limits for the transparency axis
    - alabel -- Label to place on the transparency axis
    - alines -- Add lines to indicate transparency values (e.g. p < 0.05). Can be a single number or an iterable
    - alines_colors -- Colors to use for each alpha line. Length must match alines parameter
    - alines_styles -- Line styles to use for each alpha line. Length must match alines parameter
    - cprecision -- Precision of color axis ticks
    - aprecision -- Precision of alpha axis ticks
    - black_bg -- Boolean indicating if the background to this plot is black, and hence white text/borders should be used
    - orient -- 'v' or 'h' for whether you want a vertical or horizontal colorbar
    """
    steps = 32
    if orient == 'h':
        ext = (clims[0], clims[1], alims[0], alims[1])
        cdata = np.tile(np.linspace(clims[0], clims[1], steps)[
                        np.newaxis, :], [steps, 1])
        alpha = np.tile(np.linspace(0, 1, steps)[:, np.newaxis], [1, steps])
    else:
        ext = (alims[0], alims[1], clims[0], clims[1])
        cdata = np.tile(np.linspace(clims[0], clims[1], steps)[
                        :, np.newaxis], [1, steps])
        alpha = np.tile(np.linspace(0, 1, steps)[np.newaxis, :], [steps, 1])
    color = slice_func.colorize(cdata, cm_name, clims)

    backg = np.ones((steps, steps, 3))
    acmap = slice_func.blend(backg, color, alpha)
    axes.imshow(acmap, origin='lower', interpolation='hanning',
                extent=ext, aspect='auto')

    cticks = (clims[0], np.sum(clims)/2, clims[1])
    clabels = (cfmt.format(clims[0]), clabel, cfmt.format(clims[1]))
    aticks = (alims[0], np.sum(alims)/2, alims[1])
    alabels = (afmt.format(alims[0]), alabel, afmt.format(alims[1]))

    if orient == 'h':
        axes.set_xticks(cticks)
        axes.set_xticklabels(clabels)
        axes.set_yticks(aticks)
        axes.set_yticklabels(alabels, rotation='vertical', va='center')
    else:
        axes.set_xticks(aticks)
        axes.set_xticklabels(alabels)
        axes.set_yticks(cticks)
        axes.set_yticklabels(clabels, rotation='vertical', va='center')

    if alines:
        for pos, color, style in zip(alines, alines_colors, alines_styles):
            if orient == 'h':
                axes.axhline(y=pos, linewidth=1.5,
                             linestyle=style, color=color)
            else:
                axes.axvline(x=pos, linewidth=1.5,
                             linestyle=style, color=color)

    if black_backg:
        axes.spines['bottom'].set_color('w')
        axes.spines['top'].set_color('w')
        axes.spines['right'].set_color('w')
        axes.spines['left'].set_color('w')
        axes.tick_params(axis='x', colors='w')
        axes.tick_params(axis='y', colors='w')
        axes.yaxis.label.set_color('w')
        axes.xaxis.label.set_color('w')
    else:
        axes.spines['bottom'].set_color('k')
        axes.spines['top'].set_color('k')
        axes.spines['right'].set_color('k')
        axes.spines['left'].set_color('k')
        axes.tick_params(axis='x', colors='k')
        axes.tick_params(axis='y', colors='k')
        axes.yaxis.label.set_color('k')
        axes.xaxis.label.set_color('k')
        axes.axis('on')



"""
nanslicer.py

A command-line tool for producing figures with slices through MR images. This is installed
by PIP as ``nanslicer``. Supports overlays, dual-coded overlays and colorbars. Dual-coding is
described here https://www.cell.com/neuron/fulltext/S0896-6273(12)00428-X.

The minimum command-line call is:

``nanslicer image.nii.gz output.png``

To add a dual-coded overlay, call:

``nanslicer structural.nii.gz output.png --overlay beta.nii.gz --overlay_alpha pval.nii.gz``

There are a lot of command-line options to control the colormaps and scaling. Type
``nanviewer --help`` to see a full list. The number of slices can be controlled with
the ``--slice_rows`` and ``--slice_cols`` arguments, or you can choose ``--three_axis``.
The ``--slice_axis`` and ``--slice_lims`` arguments specify the axis along which to slice,
and where to start and stop along it (expressed as fractions), for example:

``nanslicer structural.nii.gz --slice_axis x --slice_lims 0.25 0.75``

If you have timeseries data as the base image, you can plot the same slice through
each volume with ``--timeseries``, or you can choose the volume in the timeseries to use
with ``--volume N`` (the default is the first).

Controlling image quality is slightly complicated because there are two interpolation
steps. First we have to sample the 3D volumes to produce 2D slices to arbitrary
precision. Then, ``matplotlib`` has to sample those slices to plot them to the canvas.
The first step is controlled by ``--samples N``, which controls the number of points to sample in
each direction of the slice, and ``--interp_order N``, which controls the quality
of the interpolation. The defaults are 128 and 1 (linear interpolation). Increase
them to increase the quality. The ``matplotlib`` step is controlled by ``--interp METHOD``,
and can be any valid ``matplotlib` interpolation method. The default is ``hanning``,
for increased speed this can be changed to ``linear`` or ``none``. From experience,
it is the quality of the ``matplotlib`` step which is the dominant factor in figure
quality, hence the defaults of fairly fast sampling in the slicing step but using
Hanning sampling in the ``matplotlib`` step.
"""
import argparse
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from .util import add_common_arguments
from .colorbar import colorbar, alphabar
from .box import Box
from .slicer import Slicer, Axis_map
from .slice_func import scale_clip
from .layer import Layer, blend_layers


def main(args=None):
    """
    The main function that is called from the command line.

    Parameters:

    - args -- The command-line arguments. See module docstring or command-line help for a full list
    """
    parser = argparse.ArgumentParser(
        description='Takes aesthetically pleasing slices through MR images')
    add_common_arguments(parser)
    parser.add_argument('output', help='Output image name', type=str)
    parser.add_argument('--slice_rows', type=int, default=4,
                        help='Number of rows of slices')
    parser.add_argument('--slice_cols', type=int, default=5,
                        help='Number of columns of slices')
    parser.add_argument('--slice_axis', type=str, default='z',
                        help='Axis to slice along (x/y/z)')
    parser.add_argument('--slice_lims', type=float, nargs=2, default=(0.1, 0.9),
                        help='Slice between these limits along the axis, default=0.1 0.9')
    parser.add_argument('--slices', type=float, nargs='+',
                        help='Slice at specified positions')
    parser.add_argument(
        '--three_axis', help='Make a 3 axis (x,y,z) plot', action='store_true')
    parser.add_argument('--timeseries', action='store_true',
                        help='Plot the same slice through each volume in a time-series')
    parser.add_argument('--volume', type=int, default=0,
                        help='Plot one volume from a timeseries')

    parser.add_argument('--bar_pos', type=str, default='south',
                        help='Position of color-bar (south / east)')
    parser.add_argument('--figsize', type=float, nargs=2,
                        default=None, help='Figure size (width, height) in inches')
    parser.add_argument('--dpi', type=int, default=150,
                        help='DPI for output figure')
    parser.add_argument('--transpose', action='store_true',
                        help='Swap rows and columns')
    parser.add_argument('--font', type=str, default='Helvetica',
                        help='Font name, default Helvetica')
    parser.add_argument('--fontsize', type=int, default=8,
                        help='Font size, default 8')
    parser.add_argument('--title', type=str, default=None, help='Add a title')
    args = parser.parse_args()

    mpl.rc('font', family=args.font, size=args.fontsize)

    print('*** Loading base image: ', args.base_image)
    layers = [Layer(args.base_image, mask=args.mask, crop_center=args.crop_center, crop_size=args.crop_size,
                    cmap=args.base_map, clim=args.base_lims, climp=args.base_lims_p, scale=args.base_scale,
                    interp_order=args.interp_order, volume=args.volume), ]
    if args.base_lims is None:
        print('*** Base limits:', layers[0].clim)

    if args.overlay:
        layers.append(Layer(args.overlay, scale=args.overlay_scale,
                            cmap=args.overlay_map, clim=args.overlay_lim,
                            mask=args.overlay_mask, mask_threshold=args.overlay_mask_thresh,
                            alpha=args.overlay_alpha, alpha_scale=args.overlay_alpha_scale, alpha_lim=args.overlay_alpha_lim,
                            interp_order=args.interp_order))

    bbox = layers[0].bbox
    args.slice_axis = Axis_map[args.slice_axis]
    if args.three_axis:
        args.slice_rows = 1
        args.slice_cols = 3
        args.slice_axis = ['x', 'y', 'z']
        slice_total = 3
        slice_pos = (bbox.center[0], bbox.center[1], bbox.center[2])
    elif args.timeseries:
        slice_pos = bbox.center[args.slice_axis]
        slice_total = layers[0].image.shape[3]
    elif args.slices:
        slice_total = args.slice_rows*args.slice_cols
        if slice_total != len(args.slices):
            print('Number of slices and rows*cols does not match')
            exit()
        slice_pos = np.array(args.slices)
        args.slice_axis = [args.slice_axis] * slice_total
    else:
        slice_total = args.slice_rows*args.slice_cols
        slice_pos = bbox.start[args.slice_axis] + bbox.diag[args.slice_axis] * \
            np.linspace(args.slice_lims[0], args.slice_lims[1], slice_total)
        args.slice_axis = [args.slice_axis] * slice_total
    print(slice_total, ' slices in ', args.slice_rows,
          ' rows and ', args.slice_cols, ' columns')

    if args.orient == 'preclin':
        origin = 'upper'
    else:
        origin = 'lower'

    gs1 = gridspec.GridSpec(args.slice_rows, args.slice_cols)
    if not args.figsize:
        args.figsize = (3*args.slice_cols, 3*args.slice_rows)
    figure = plt.figure(facecolor='black', figsize=args.figsize)

    print('*** Slicing')
    for s in range(0, slice_total):
        if args.transpose:
            col, row = divmod(s, args.slice_rows)
        else:
            row, col = divmod(s, args.slice_cols)
        ax = plt.subplot(gs1[row, col], facecolor='black')
        if args.timeseries:
            layers[0].volume = s
            sp = slice_pos
            axis = args.slice_axis
        else:
            sp = slice_pos[s]
            axis = args.slice_axis[s]

        slcr = Slicer(bbox, sp, axis, args.samples, orient=args.orient)
        sl_final = blend_layers(layers, slcr)
        ax.imshow(sl_final, origin=origin, extent=slcr.extent,
                  interpolation=args.interp)
        ax.axis('off')
        if args.contour:
            sl_contour = layers[1].get_alpha(slcr)
            contour_levels = scale_clip(
                np.array(args.contour), args.overlay_alpha_lim)

            # Contour levels must be within the range of overlay alpha values.
            # Ignore contour levels that are not within this range to prevent
            # spurious contour lines from being drawn.
            valid_levels = (np.min(sl_contour) < contour_levels) & (
                contour_levels < np.max(sl_contour))
            if any(valid_levels):
                ax.contour(sl_contour, levels=contour_levels[valid_levels], origin=origin, extent=slcr.extent,
                           colors=args.contour_color, linestyles=args.contour_style, linewidths=1)

    if args.base_label or args.overlay_label:
        print('*** Adding colorbar')
        if args.bar_pos.lower() == 'south':
            cbar_bottom = 0.3 * (args.fontsize / 12) / args.figsize[1]
            cbar_top = cbar_bottom + 0.1 / args.figsize[1]
            gs1.update(left=0.01, right=0.99, bottom=cbar_top+0.001,
                       top=0.99, wspace=0.01, hspace=0.01)
            gs2 = gridspec.GridSpec(1, 1)
            gs2.update(left=0.05, right=0.95, bottom=cbar_bottom,
                       top=cbar_top, wspace=0.1, hspace=0.1)
            c_orient = 'h'
            c_axes = plt.subplot(gs2[0], facecolor='black')
        elif args.bar_pos.lower() == 'south_out':
            gs1.update(left=0.01, right=0.99, bottom=0.01,
                       top=0.99, wspace=0.01, hspace=0.01)
            c_orient = 'h'
            c_axes = figure.add_subplot(3, 3, 8)
            c_axes.set_position([0.1, 0.1, 0.8, 0.05])
            print('Rect: ', c_axes.get_position())
        elif args.bar_pos.lower() == 'east':
            cbarw = 0.275 * (args.fontsize / 12) / args.figsize[0]
            gs1.update(left=0.01, right=1 - cbarw, bottom=0.01,
                       top=0.99, wspace=0.01, hspace=0.01)
            gs2 = gridspec.GridSpec(1, 1)
            gs2.update(left=1 - cbarw + 0.001, right=1 - cbarw/1.5, bottom=0.08,
                       top=0.92, wspace=0.01, hspace=0.01)
            c_orient = 'v'
            c_axes = plt.subplot(gs2[0], facecolor='black')
        else:
            raise ValueError('Unsupported bar position: ' + args.bar_pos)

        if args.overlay_alpha:
            alphabar(c_axes, args.overlay_map, args.overlay_lim, args.overlay_label,
                     args.overlay_alpha_lim, args.overlay_alpha_label, orient=c_orient)
        else:
            if args.base_map:
                colorbar(c_axes, layers[0].cmap, layers[0].clim,
                         args.base_label, orient=c_orient)
            else:
                colorbar(c_axes, layers[1].cmap, layers[1].clim,
                         args.overlay_label, orient=c_orient)
    else:
        gs1.update(left=0.01, right=0.99, bottom=0.01,
                   top=0.99, wspace=0.01, hspace=0.01)

    if args.title:
        figure.axes[-1].text(0.01, 0.99, args.title, color='w', size=args.fontsize, verticalalignment='top',
                             transform=figure.transFigure)
    print('*** Saving')
    print('Writing file: ', args.output, 'at', args.dpi, ' DPI')
    figure.savefig(args.output, facecolor=figure.get_facecolor(),
                   edgecolor='none', dpi=args.dpi)
    plt.close(figure)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
slicer.py

This module contains the core Slicer object that samples Layers to produce
image arrays that can be drawn with matlplotlib.
"""
import numpy as np
import scipy.ndimage.interpolation as ndinterp

Axis_map = {'x': 0, 'y': 1, 'z': 2}
Orient_map = {'clin': ({0: 1, 1: 0, 2: 0}, {0: 2, 1: 2, 2: 1}),
              'preclin': ({0: 2, 1: 2, 2: 0}, {0: 1, 1: 0, 2: 1})}


def axis_indices(axis, orient='clin'):
    """
    Returns a pair of indices corresponding to right/up for the given orientation.
    Parameters:
    axis:   The perpendicular axis to the slice. Use Axis_map to convert between x/y/z and 0/1/2
    orient: Either 'clin' or 'preclin'
    """
    this_orient = Orient_map[orient]
    return (this_orient[0][axis], this_orient[1][axis])


class Slicer:
    """
    The Slicer class.

    When constructed, creates an array of world-space co-ordinates which represent the desired slice

    Constructor Parameters:

    - bbox -- :py:class:`~nanslice.box.Box` instance that defines the world-space bounding box that you want to slice
    - pos -- Position within the box to generate the slice through
    - axis -- Which axis you want to slice across. Either x/y/z or 0/1/2
    - samples -- Number of samples in the horizontal direction
    - orient --  'clin' or 'preclin'
    """

    def __init__(self, bbox, pos, axis, samples=64, orient='clin'):
        try:
            ind_0 = Axis_map[axis]  # If someone passed in x/y/z
        except KeyError:
            ind_0 = axis  # Assume it was an integer

        ind_1, ind_2 = axis_indices(ind_0, orient=orient)
        start = np.copy(bbox.start)
        start[ind_0] = pos
        dir_rt = np.zeros((3,))
        dir_up = np.zeros((3,))
        dir_rt[ind_1] = bbox.diag[ind_1]
        dir_up[ind_2] = bbox.diag[ind_2]
        aspect = np.linalg.norm(dir_up) / np.linalg.norm(dir_rt)
        samples_up = int(round(aspect * samples))
        self._world_space = (start[:, None, None] +
                             dir_rt[:, None, None] * np.linspace(0, 1, samples)[None, :, None] +
                             dir_up[:, None, None] * np.linspace(0, 1, samples_up)[None, None, :])
        # This is the extent parameter for matplotlib
        self.extent = (bbox.start[ind_1], bbox.end[ind_1],
                       bbox.start[ind_2], bbox.end[ind_2])
        self._tfm = None
        self._voxel_space = None

    def get_voxel_coords(self, tfm):
        """
        Returns an array of voxel space co-ordinates for this slice, which will be cached.
        If a subsequent call uses the same affine transform, the cached co-ordinates will be returned.
        If a new transform is passed in, then a fresh set of co-ordinates are calculated first.

        Parameters:

        - tfm -- An affine transform that defines an images physical space (usually the .affine property of an nibabel image)
        """
        if not np.array_equal(tfm, self._tfm):
            old_sz = self._world_space.shape
            new_sz = np.prod(self._world_space.shape[1:])
            scale = np.mat(tfm[0:3, 0:3]).I
            offset = np.dot(-scale, tfm[0:3, 3]).T
            isl = np.dot(scale, self._world_space.reshape(
                [3, new_sz])) + offset[:]
            isl = np.array(isl).reshape(old_sz)
            self._voxel_space = isl
            self._tfm = tfm
        return self._voxel_space

    def sample(self, img_data, affine, order, scale=1.0, volume=0):
        """
        Samples the passed 3D/4D image and returns a 2D slice

        Paramters:

        - img -- An nibabel image
        - order -- Interpolation order. 1 is linear interpolation
        - scale -- Scale factor to multiply all voxel values by
        - volume -- If sampling 4D data, specify the desired volume

        """
        physical = self.get_voxel_coords(affine)
        # Support timeseries by adding an extra co-ord specifying the volume
        if len(img_data.shape) == 4:
            vol_index = np.tile(volume, physical.shape[1:3])[np.newaxis, :]
            physical = np.concatenate((physical, vol_index), axis=0)
        return scale * ndinterp.map_coordinates(img_data, physical, order=order).T



#!/usr/bin/env python
"""
The nanslice package. Everything important is in a submodule, but the
:py:class:`~nanslice.slicer.Slicer` and :py:class:`~nanslice.layer.Layer` classes,
and the :py:func:`~nanslice.layer.blend_layers` function, are imported here
for convenient access. Between them, they do all the important work.
"""

from .layer import Layer, blend_layers
from .slicer import Slicer



#!/usr/bin/env python
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
import ipywidgets as ipy
from . import util
from .slicer import Slicer, Axis_map
from .layer import Layer, blend_layers
from .colorbar import colorbar, alphabar


def three_plane(images, orient='clin', samples=128,
                cbar=None, contour=None,
                interactive=False):
    """
    Draw a standard 3-plane view through the center of the image

    Parameters:

    - images -- Either a filename (in a string or Path object) or a list of Layer objects
    - orient -- 'clin' or 'preclin'
    - sample -- Number of samples, default is 128, higher is better but slower
    - cbar -- Adds a colorbar. Either 'True' or an integer giving the layer the colorbar is from
    - contour -- Value(s) that contours are drawn at. Requires that cbar be set - contours will come from same layer
    - interactive -- Add sliders and make the view interactive
    """


def three_plane(images, orient='clin', samples=128,
                volume=0, cbar=None, contour=None,
                interactive=False, title=None):
    if interactive:
        plt.ion()
    else:
        plt.ioff()
    if isinstance(images, str):
        layers = [Layer(images, volume=volume), ]
    elif isinstance(images, Layer):
        layers = [images, ]
    elif isinstance(images[0], str):
        layers = [Layer(img, volume=volume) for img in images]
    else:
        layers = images
    bbox = layers[0].bbox
    gs1 = gs.GridSpec(1, 3)
    fig = plt.figure(facecolor='black', figsize=(9, 3))
    if cbar:
        if isinstance(cbar, bool):
            cbar = 0
        gs1.update(left=0.01, right=0.88, bottom=0.01,
                   top=0.99, wspace=0.01, hspace=0.01)
        gs2 = gs.GridSpec(1, 1)
        gs2.update(left=0.92, right=0.98, bottom=0.1,
                   top=0.9, wspace=0.1, hspace=0.1)
        cax = fig.add_subplot(gs2[0], facecolor='white')
        clayer = layers[cbar]
        if clayer.alpha_image:
            if contour:
                try:
                    iterator = iter(contour)
                except:
                    contour = (contour,)
            alphabar(cax,
                     clayer.cmap, clayer.clim, clayer.label,
                     clayer.alpha_lim, clayer.alpha_label,
                     alines=contour, black_backg=True, orient='v')
        else:
            colorbar(cax,
                     clayer.cmap, clayer.clim, clayer.label,
                     black_backg=True, orient='v')
    else:
        gs1.update(left=0.01, right=0.99, bottom=0.01,
                   top=0.99, wspace=0.01, hspace=0.01)
    implots = [None, None, None]
    iax = [None, None, None]

    def wrap_sections(pos_x, pos_y, pos_z):
        pos = (pos_x, pos_y, pos_z)
        for i in range(3):
            slcr = Slicer(bbox, pos[i], i, samples=samples, orient=orient)
            blended_slice = blend_layers(layers, slcr)
            if implots[i]:
                implots[i].set_data(blended_slice)
            else:
                iax[i] = fig.add_subplot(gs1[i], facecolor='black')
                implots[i] = iax[i].imshow(
                    blended_slice, origin='lower', extent=slcr.extent, interpolation='nearest')
                iax[i].axis('off')
            if contour:
                sl_contour = layers[cbar].get_alpha(slcr)
                iax[i].contour(sl_contour, levels=contour, origin='lower', extent=slcr.extent,
                               colors='k', linestyles='-', linewidths=1)
    wrap_sections(bbox.center[0], bbox.center[1], bbox.center[2])
    if title:
        fig.suptitle(title, color='white')
    if interactive:
        # Setup widgets
        slider_x = ipy.FloatSlider(min=bbox.start[0], max=bbox.end[0], value=bbox.center[0],
                                   continuous_update=True, description='X:')
        slider_y = ipy.FloatSlider(min=bbox.start[1], max=bbox.end[1], value=bbox.center[1],
                                   continuous_update=True, description='Y:')
        slider_z = ipy.FloatSlider(min=bbox.start[2], max=bbox.end[2], value=bbox.center[2],
                                   continuous_update=True, description='Z:')
        widgets = ipy.interactive(
            wrap_sections, pos_x=slider_x, pos_y=slider_y, pos_z=slider_z)
        # Now do some manual layout
        # Set the sliders to horizontal layout
        hbox = ipy.HBox(widgets.children[0:3])
        vbox = ipy.VBox((hbox, widgets.children[3]))
        # iplot.widget.children[-1].layout.height = '350px'
        return vbox
    else:
        plt.close()
        return fig


def slice_axis(images, nrows=1, ncols=3, slice_axis='z', slice_lims=(0.25, 0.75),
               orient='clin', samples=128,
               cbar=None, contour=None, title=None):
    ntotal = nrows*ncols
    slice_axes = slice_axis * ntotal
    slice_pos = np.linspace(slice_lims[0], slice_lims[1], num=ntotal)
    return slices(images, nrows, ncols, slice_axes, slice_pos, False, orient, samples, cbar, contour, title)


def slices(images, nrows=1, ncols=1, slice_axes=['z', ], slice_pos=[0.5, ], absolute=False,
           orient='clin', samples=128,
           cbar=None, contour=None, title=None):
    if isinstance(images, str):
        layers = [Layer(images), ]
    elif isinstance(images, Layer):
        layers = [images, ]
    elif isinstance(images[0], str):
        layers = [Layer(img) for img in images]
    else:
        layers = images
    plt.ioff()
    bbox = layers[0].bbox
    gs1 = gs.GridSpec(nrows, ncols)
    fig = plt.figure(facecolor='black', figsize=(3*ncols, 3*nrows))
    if cbar:
        if isinstance(cbar, bool):
            cbar = 0
        gs1.update(left=0.01, right=0.88, bottom=0.01,
                   top=0.99, wspace=0.01, hspace=0.01)
        gs2 = gs.GridSpec(1, 1)
        gs2.update(left=0.92, right=0.98, bottom=0.1,
                   top=0.9, wspace=0.1, hspace=0.1)
        cax = fig.add_subplot(gs2[0], facecolor='white')
        clayer = layers[cbar]
        if clayer.alpha_image:
            if contour:
                try:
                    iterator = iter(contour)
                except:
                    contour = (contour,)
            alphabar(cax,
                     clayer.cmap, clayer.clim, clayer.label,
                     clayer.alpha_lim, clayer.alpha_label,
                     alines=contour, black_backg=True, orient='v')
        else:
            colorbar(cax,
                     clayer.cmap, clayer.clim, clayer.label,
                     black_backg=True, orient='v')
    else:
        gs1.update(left=0.01, right=0.99, bottom=0.01,
                   top=0.99, wspace=0.01, hspace=0.01)
    for row in range(nrows):
        for col in range(ncols):
            i = row*ncols + col
            axis = slice_axes[i]
            if absolute:
                pos = slice_pos[i]
            else:
                pos = bbox.start[Axis_map[axis]] + \
                    bbox.diag[Axis_map[axis]]*slice_pos[i]
            slcr = Slicer(bbox, pos, axis, samples=samples, orient=orient)
            blended_slice = blend_layers(layers, slcr)
            iax = fig.add_subplot(gs1[row, col], facecolor='black')
            iax.imshow(blended_slice, origin='lower',
                       extent=slcr.extent, interpolation='nearest')
            iax.axis('off')
            if contour:
                sl_contour = layers[cbar].get_alpha(slcr)
                iax.contour(sl_contour, levels=contour, origin='lower', extent=slcr.extent,
                            colors='k', linestyles='-', linewidths=1)
    if title:
        fig.suptitle(title, color='white')
    plt.close()
    return fig


def timeseries(image, axis='z', orient='clin', clim=None, title=None):
    series = Layer(image, clim=clim)
    plt.ioff()
    bbox = series.bbox
    N = series.matrix[3]
    gs1 = gs.GridSpec(1, N)
    fig = plt.figure(facecolor='black', figsize=(3*N, 3))

    ax_ind = Axis_map[axis]
    slcr = Slicer(bbox, bbox.center[ax_ind],
                  axis, samples=series.matrix[ax_ind], orient=orient)
    for i in range(N):
        series.volume = i
        sl = series.get_color(slcr)
        iax = fig.add_subplot(gs1[i], facecolor='black')
        iax.imshow(sl, origin='lower', extent=slcr.extent,
                   interpolation='nearest')
        iax.axis('off')
    if title:
        plt.suptitle(title, color='w')
    plt.close()
    return fig


def compare(image1, image2, axis='z', orient='clin', samples=128,
            clim=None, title=None):
    layer1 = Layer(image1, interp_order=0, clim=clim)
    layer2 = Layer(image2, interp_order=0, clim=layer1.clim)
    plt.ioff()
    bbox = layer1.bbox
    gs1 = gs.GridSpec(1, 2)
    fig = plt.figure(facecolor='black', figsize=(6, 3))

    gs1.update(left=0.01, right=0.88, bottom=0.01,
               top=0.99, wspace=0.01, hspace=0.01)
    gs2 = gs.GridSpec(1, 1)
    gs2.update(left=0.92, right=0.98, bottom=0.1,
               top=0.9, wspace=0.1, hspace=0.1)
    cax = fig.add_subplot(gs2[0], facecolor='white')
    colorbar(cax,
             layer1.cmap, layer1.clim, layer1.label,
             black_backg=True, orient='v')

    ax_ind = Axis_map[axis]
    slcr = Slicer(bbox, bbox.center[ax_ind],
                  axis, samples=layer1.matrix[ax_ind], orient=orient)
    slice1 = layer1.get_color(slcr)
    slice2 = layer2.get_color(slcr)
    iax = fig.add_subplot(gs1[0], facecolor='black')
    iax.imshow(slice1, origin='lower', extent=slcr.extent,
               interpolation='nearest')
    iax.axis('off')
    iax = fig.add_subplot(gs1[1], facecolor='black')
    iax.imshow(slice2, origin='lower', extent=slcr.extent,
               interpolation='nearest')
    iax.axis('off')
    if title:
        fig.suptitle(title, color='white')
    plt.close()
    return fig



#!/usr/bin/env python
"""slice_func.py

Functions for manipulating 'slices'/images (or (X, Y, 3) arrays)
"""

import numpy as np
import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.colors as colors
import scipy.ndimage.filters as filters
import colorcet as cc


class MidNorm(mpl.colors.Normalize):
    def __init__(self, vmin=None, vmax=None, clip=False):
        if vmin < 0:
            self.midpoint = 0
        else:
            self.midpoint = None
        mpl.colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        result, is_scalar = self.process_value(value)
        self.autoscale_None(result)

        if self.midpoint:
            x, y = [-self.vmax, self.midpoint, self.vmax], [0, 0.5, 1]
            return np.ma.masked_array(np.interp(value, x, y))
        else:
            return mpl.colors.Normalize.__call__(value, clip)


def colorize(data, cmap, clims):
    """
    Apply a colormap to grayscale data. Takes an (X, Y) array and returns an (X, Y, 3) array

    Parameters:

    - data -- The 2D scalar (X, Y) array to colorize
    - cmap -- Any valid matplotlib colormap or colormap name
    - clims -- The limits for the colormap
    """
    if cmap == 'twoway':
        c_neg = cm.get_cmap('cet_CET_L15')
        c_plus = cm.get_cmap('cet_CET_L3')
        cmap = colors.LinearSegmentedColormap.from_list(
            'twoway', np.vstack((c_neg(np.linspace(1, 0, 128)),
                                 c_plus(np.linspace(0, 1, 128)))))
        norm = colors.TwoSlopeNorm(vmin=clims[0], vcenter=0, vmax=clims[1])
    elif cmap == 'phase':
        cmap = cc.m_colorwheel
        norm = colors.Normalize(vmin=clims[0], vmax=clims[1])
    else:
        cmap = mpl.cm.get_cmap(cmap)
        norm = colors.Normalize(vmin=clims[0], vmax=clims[1])
    smap = mpl.cm.ScalarMappable(norm=norm, cmap=cmap)
    return smap.to_rgba(data, alpha=1, bytes=False)[:, :, 0: 3]


def scale_clip(data, lims):
    """
    Scale an image to fill the range 0-1  and clip values that fall outside that range

    Parameters:

    - data -- The image data array
    - lims -- The limits to scale betwee
    """
    return np.clip((data - lims[0]) / (lims[1] - lims[0]), 0, 1)


def blend(img_under, img_over, img_alpha):
    """
    Blend together two images using an alpha channel image

    Parameters:

    - img_under -- The base image (underneath the overlay)
    - img_over  -- The overlay image
    - img_alpha -- Transparency/alpha value to use when blending
    """
    return img_under*(1 - img_alpha[:, :, None]) + img_over*img_alpha[:, :, None]


def mask(img, img_mask, back=np.array((0, 0, 0))):
    """
    Mask out sections of one image using another

    Parameters:

    - img -- The image to be masked
    - img_mask -- The mask image
    - back -- Background value
    """
    if img_mask is None:
        return img
    if back.ndim == 1:
        masked = np.where(img_mask[:, :, np.newaxis],
                          img, back[np.newaxis, np.newaxis, :])
    elif back.ndim == 2:
        masked = np.where(img_mask[:, :, np.newaxis],
                          img, back[:, :, np.newaxis])
    elif back.ndim == 3:
        masked = np.where(img_mask[:, :, np.newaxis], img, back)
    else:
        raise Exception(
            'Masking requires a 1, 2, or 3 dimensional array as the background')
    return masked


def blur(img, sigma=1):
    """
    Blur an image with a Gaussian kernel

    Parameters:

    - img -- The image to blur
    - sigma -- The FWHM of the Gaussian kernel, in voxels
    """
    return filters.gaussian_filter(img, sigma)


def checkerboard(img1, img2, square_size=16):
    """Combine two images in a checkerboard pattern, useful for checking image
       registration quality. Idea stolen from @ramaana_ on Twitter"""
    if (img1.shape != img2.shape):
        raise Exception('Image shape do not match:' +
                        str(img1.shape) + ' vs:' + str(img2.shape))
    shape = img1.shape
    img3 = np.zeros_like(img1)
    from1 = True
    row = 0
    row_sz = square_size
    while row < shape[0]:
        col = 0
        col_sz = square_size
        while col < shape[1]:
            if from1:
                img3[row:row+row_sz, col:col+col_sz,
                     :] = img1[row:row+row_sz, col:col+col_sz, :]
            else:
                img3[row:row+row_sz, col:col+col_sz,
                     :] = img2[row:row+row_sz, col:col+col_sz, :]
            col = col + col_sz
            if (col + col_sz) > shape[1]:
                col_sz = shape[1] - col
            from1 = not from1
        row = row + row_sz
        if (row + row_sz) > shape[0]:
            row_sz = shape[0] - row
        from1 = not from1
    return img3



from os import path, getcwd
from .layer import Layer, blend_layers
from .slice_func import scale_clip
from .slicer import Slicer, Axis_map
from .box import Box
from .colorbar import colorbar, alphabar
from .util import add_common_arguments
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
from nipype.interfaces.base import BaseInterface, BaseInterfaceInputSpec, traits, TraitedSpec, isdefined


class SlicesInputSpec(BaseInterfaceInputSpec):
    out_file = traits.File(mandatory=True, desc='Output slice image')
    base_file = traits.File(exists=True, mandatory=True,
                            desc='Image file to slice')
    base_map = traits.String('gist_gray', usedefault=True,
                             desc='Color map for base image')
    base_window = traits.Tuple(
        minlen=2, maxlen=2, desc='Window for base image')
    base_scale = traits.Float(1.0, usedefault=True,
                              desc='Scaling factor for base image')
    base_label = traits.String(
        '', usedefault=True, desc='Label for base image color-bar')

    mask_file = traits.File(None, usedefault=True,
                            mandatory=True, desc='Mask file')

    slice_axis = traits.String('z', usedefault=True, desc='Axis to slice')
    slice_lims = traits.Tuple((0.1, 0.9), minlen=2, maxlen=2, usedefault=True,
                              desc='Limits of axis to slice, fraction (low, high) default (0.1, 0.9)')
    slice_layout = traits.Tuple(
        (3, 6), minlen=2, maxlen=2, usedefault=True, desc='Slices layout (rows, cols)')
    volume = traits.Int(0, usedefault=True,
                        desc='Volume for slicing in multi-volume file')
    figsize = traits.Tuple(minlen=2, maxlen=2, desc='Output figure size')
    preclinical = traits.Bool(False, usedefault=True,
                              desc='Data is pre-clinical, fix orientation')
    transpose = traits.Bool(False, usedefault=True,
                            desc='Transpose slice layout')
    bar_pos = traits.String('bottom', usedefault=True,
                            desc='Color/Alphabar position')


class SlicesOutputSpec(TraitedSpec):
    out_file = traits.File(desc="Slice image")


class Slices(BaseInterface):
    input_spec = SlicesInputSpec
    output_spec = SlicesOutputSpec

    def _run_interface(self, runtime):
        inputs = self.inputs
        print('Loading base image: ', inputs.base_file)
        if isdefined(inputs.base_window):
            base_window = inputs.base_window
        else:
            base_window = None
        layers = [Layer(inputs.base_file, mask=inputs.mask_file,
                        cmap=inputs.base_map, clim=base_window, scale=inputs.base_scale,
                        volume=inputs.volume), ]

        bbox = layers[0].bbox
        slice_axis = Axis_map[inputs.slice_axis]

        slice_total = inputs.slice_layout[0]*inputs.slice_layout[1]
        slice_pos = bbox.start[slice_axis] + bbox.diag[slice_axis] * \
            np.linspace(inputs.slice_lims[0],
                        inputs.slice_lims[1], slice_total)
        slice_axis = [slice_axis] * slice_total

        if inputs.preclinical:
            origin = 'upper'
            orient = 'preclin'
        else:
            origin = 'lower'
            orient = 'clin'

        gs1 = gridspec.GridSpec(*inputs.slice_layout)
        if isdefined(inputs.figsize):
            f = plt.figure(facecolor='black', figsize=inputs.figsize)
        else:
            f = plt.figure(facecolor='black', figsize=(
                3*inputs.slice_layout[0], 3*inputs.slice_layout[1] + 1))

        for s in range(0, slice_total):
            if inputs.transpose:
                col, row = divmod(s, inputs.slice_layout[0])
            else:
                row, col = divmod(s, inputs.slice_layout[1])
            ax = plt.subplot(gs1[row, col], facecolor='black')

            slcr = Slicer(bbox, slice_pos[s],
                          slice_axis[s], 256, orient=orient)
            sl_final = blend_layers(layers, slcr)
            ax.imshow(sl_final, origin=origin, extent=slcr.extent)
            ax.axis('off')

        if isdefined(inputs.base_label):
            print('*** Adding colorbar')
            if inputs.bar_pos == 'bottom':
                gs1.update(left=0.01, right=0.99, bottom=(0.4 / (1 + inputs.slice_layout[0])),
                           top=0.99, wspace=0.01, hspace=0.01)
                gs2 = gridspec.GridSpec(1, 1)
                gs2.update(left=0.08, right=0.92, bottom=(0.2 / (1 + inputs.slice_layout[1])),
                           top=(0.38 / (1 + inputs.slice_layout[0])), wspace=0.1, hspace=0.1)
                orient = 'h'
            else:
                gs1.update(left=0.01, right=0.95, bottom=0.01,
                           top=0.99, wspace=0.01, hspace=0.01)
                gs2 = gridspec.GridSpec(1, 1)
                gs2.update(left=0.97, right=0.99, bottom=0.05,
                           top=0.95, wspace=0.01, hspace=0.01)
                orient = 'v'
            axes = plt.subplot(gs2[0], facecolor='black')
            if inputs.base_map:
                colorbar(axes, layers[0].cmap, layers[0].clim,
                         inputs.base_label, orient=orient)
        else:
            gs1.update(left=0.01, right=0.99, bottom=0.01,
                       top=0.99, wspace=0.01, hspace=0.01)
        print('*** Saving')
        print('Writing file: ', inputs.out_file)
        f.savefig(inputs.out_file, facecolor=f.get_facecolor(),
                  edgecolor='none')
        plt.close(f)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = path.abspath(self.inputs.out_file)
        print('Out file:', outputs['out_file'])
        return outputs



#!/usr/bin/env python
"""Box.py

Contains a simple bounding-box class"""

import numpy as np


class Box:
    """
    A simple bounding-box class

    Constructor parameters:

    - center -- The center of the box (x,y,z) in world-space co-ordinates
    - size   -- The size of the box (x,y,z) in world-space units
    - corners -- Two corners (x,y,z) of the box in world-space co-ordinates

    Either corners or both center and size must be specified
    """

    def __init__(self, center=None, size=None, corners=None):
        if center is not None and size is not None:
            self._center = np.array(center)
            self._diag = np.array(size)
            self._c = (self._center - self._diag/2,
                       self._center + self._diag/2)
        elif corners is not None:
            self._c = corners
            self._diag = self._c[1] - self._c[0]
            self._center = (self._c[0] + self._c[1]) / 2
        else:
            raise Exception('Either center & sz or corners must be specified')

    def __str__(self):
        return 'Box Start: ' + str(self.start) + ' End: ' + str(self.end)

    @classmethod
    def fromImage(cls, img):
        """
        Creates a bounding box from the corners defined by an image

        Parameters:

        - img -- An nibabel image
        """
        img_shape = img.get_data().shape
        corners = np.array([[0, 0, 0, 1.],
                            [img_shape[0] - 1, img_shape[1] - 1, img_shape[2] - 1, 1.]])
        corners = np.dot(img.get_affine(), corners.T)
        corner1 = np.min(corners[0:3, :], axis=1)
        corner2 = np.max(corners[0:3, :], axis=1)
        return cls(corners=(corner1, corner2))

    @classmethod
    def fromMask(cls, img, padding=0):
        """
        Creates a bounding box that encloses all non-zero voxels in a volume

        Parameters:

        - img -- The volume to create the bounding-box from
        - padding -- Number of extra voxels to pad the resulting box by
        """
        data = img.get_data()

        # Individual axis min/maxes
        xmin, xmax = np.where(np.any(data, axis=(1, 2)))[0][[0, -1]]
        ymin, ymax = np.where(np.any(data, axis=(0, 2)))[0][[0, -1]]
        zmin, zmax = np.where(np.any(data, axis=(0, 1)))[0][[0, -1]]

        # Convedir_rt to physical space
        corners = np.array([[xmin, ymin, zmin, 1.],
                            [xmax, ymax, zmax, 1.]])
        corners = np.dot(img.get_affine(), corners.T)
        # Now do min/maxes again to standardise corners
        corner1 = np.min(corners[0:3, :], axis=1) - padding
        corner2 = np.max(corners[0:3, :], axis=1) + padding
        return cls(corners=(corner1, corner2))

    @property
    def start(self):
        """Returns the 'start' corner of the bounding-box"""
        return self._c[0]

    @property
    def end(self):
        """Returns the 'end' corner of the bounding-box (opposite the start)"""
        return self._c[1]

    @property
    def diag(self):
        """Returns the vector diagonal of the bounding-box"""
        return self._diag

    @property
    def center(self):
        """Returns the geometric center of the bounding-box"""
        return self._center

    def slice_positions(self, num_slices, start=0, end=1):
        """Returns an array of slice positions along the specified axis"""
        slice_pos = self.start + self.diag * \
            np.linspace(start, end, num_slices)[:, np.newaxis]
        return slice_pos



#!/usr/bin/env python
"""layer.py

Contains the :py:class:`~nanslice.layer.Layer` class and the :py:func:`~nanslice.layer.blend_layers`
function.
"""
from numpy import isfinite, nanpercentile, ma, ones_like, array
from nibabel import load
from . import slice_func
from .box import Box
from .util import ensure_image, check_path


class Layer:
    """
    The Layer class

    Each layer consists of a base MR image, with optional mask and alpha (transparency) images
    and their associated parameters (colormap, limits, scales etc.)

    Constructor parameters:

    - image -- The image contained in this layer. Can be either a string/path to an image file or an nibabel image
    - scale  -- A scaling factor to multiply all voxels in the image by
    - volume -- If reading a 4D file, specify which volume to use
    - interp_order -- Interpolation order. 1 is linear interpolation
    - cmap  -- The colormap to apply to the Layer. Any valid matplotlib colormap
    - clim  -- The limits (min, max) values to use for the colormap
    - label -- The label for this layer (used for colorbars)
    - mask           -- A mask image to use with this layer
    - mask_threshold -- Apply a threshold (lower) to the mask
    - crop_center -- Center of box to crop to
    - crop_size   -- Size of crop box
    - alpha       -- An alpha (transparency) image to use with this layer
    - alpha_lim  -- Specify the limits/window for the alpha image
    - alpha_scale -- Scaling factor for the alpha image
    - alpha_label -- Label for the alpha axis on alphabars
    - background -- Background color for masking, either 'black' (default) or 'white'

    """

    def __init__(self, image, scale=1.0, volume=0, interp_order=1,
                 cmap=None, clim=None, climp=None, label='',
                 mask=None, mask_threshold=0, crop_center=None, crop_size=None,
                 alpha=None, alpha_lim=None, alpha_scale=1.0, alpha_label='',
                 background='black'):
        self.image = ensure_image(image)
        self.scale = scale
        self.interp_order = interp_order
        self.volume = volume
        self.label = label

        self.img_data = self.image.get_data()
        self.img_data[~isfinite(self.img_data)] = 0
        self.matrix = self.img_data.shape

        self.mask_image = ensure_image(mask)
        self.mask_threshold = mask_threshold
        if crop_center and crop_size:
            self.bbox = Box(center=crop_center, size=crop_size)
        elif self.mask_image:
            self.bbox = Box.fromMask(self.mask_image)
        else:
            self.bbox = Box.fromImage(self.image)

        if cmap:
            self.cmap = cmap
        else:
            self.cmap = 'gist_gray'

        if clim is not None:
            self.clim = clim
        else:
            if len(self.img_data.shape) == 4:
                limdata = self.img_data[:, :, :, self.volume].squeeze()
            else:
                limdata = self.img_data
            if self.mask_image:
                limdata = ma.masked_where(
                    self.mask_image.get_data() == 0, limdata).compressed()
            if climp is None:
                climp = (2, 98)
            self.clim = nanpercentile(limdata, climp)

        if check_path(alpha):
            self.alpha_image = load(str(alpha))
            if alpha_lim:
                self.alpha_lim = alpha_lim
            else:
                self.alpha_lim = nanpercentile(
                    self.alpha_image.get_data(), (2, 98))
        elif alpha:
            self.alpha_image = ones_like(self.image) * alpha
        else:
            self.alpha_image = None
        self.alpha_label = alpha_label
        self.alpha_scale = alpha_scale

        if background == 'white':
            self._back = array([1])
        else:
            self._back = array([0])

    def get_slice(self, slicer):
        """
        Returns a slice through the base image

        Parameters:

        - slicer -- The :py:class:`~nanslice.slicer.Slicer` object to slice this layer with
        """
        return slicer.sample(self.img_data, self.image.affine, self.interp_order, self.scale, self.volume)

    def get_color(self, slicer):
        """
        Returns a colorized slice through the base image contained in the Layer

        Parameters:

        - slicer -- The :py:class:`~nanslice.slicer.Slicer` object to slice this layer with
        """
        return slice_func.colorize(self.get_slice(slicer), self.cmap, self.clim)

    def get_mask(self, slicer):
        if self.mask_image:
            mask_slc = slicer.sample(self.mask_image.get_data(
            ), self.mask_image.affine, 0) > self.mask_threshold
        elif self.mask_threshold:
            mask_slc = slicer.sample(
                self.img_data, self.image.affine, self.interp_order, self.scale, self.volume) > self.mask_threshold
        else:
            return None
        return mask_slc

    def get_alpha(self, slicer):
        """
        Returns the alpha (transparency) slice for this Layer

        Parameters:

        - slicer -- The :py:class:`~nanslice.slicer.Slicer` object to slice this layer with
        """

        if self.alpha_image:
            alpha_slice = slicer.sample(
                self.alpha_image.get_data(), self.alpha_image.affine, self.interp_order, self.alpha_scale)
            alpha_slice = slice_func.scale_clip(alpha_slice, self.alpha_lim)
            return alpha_slice
        else:
            return None

    def plot(self, slicer, axes):
        """
        Plot a Layer into a Matplotlib axes using the provided Slicer

        Parameters:

        - slicer -- The :py:class:`~nanslice.slicer.Slicer` object to slice this layer with
        - axes   -- A matplotlib axes object
        """
        slc = slice_func.mask(self.get_color(
            slicer), self.get_mask(slicer), back=self._back)
        cax = axes.imshow(slc, origin='lower',
                          extent=slicer.extent, interpolation='nearest')
        axes.axis('off')
        return cax


def blend_layers(layers, slicer):
    """
    Blends together a set of overlays using their alpha information

    Parameters:

    - layers -- An iterable (e.g. list/tuple) of :py:class:`Layer` objects
    - slicer -- The :py:class:`~nanslice.slicer.Slicer` object to slice the layers with    
    """
    slc = slice_func.mask(layers[0].get_color(
        slicer), layers[0].get_mask(slicer))
    for next_layer in layers[1:]:
        next_slc = next_layer.get_color(slicer)
        if next_layer.alpha_image:
            next_alpha = next_layer.get_alpha(slicer)
            slc = slice_func.blend(slc, next_slc, next_alpha)
        else:
            slc = slice_func.mask(next_slc, next_layer.get_mask(slicer), slc)
    return slc



#!/usr/bin/env python
"""nanviewer.py

A simple viewer to demonstrate 'dual-coded' neuroimaging overlays. This is installed by PIP
as ``nanviewer``.

Dual-coding is described here https://www.cell.com/neuron/fulltext/S0896-6273(12)00428-X.
Python code adapted from http://matplotlib.org/examples/user_interfaces/embedding_in_qt5.html.

The viewer can be started on the command-line by calling:

``nanviewer image.nii.gz``

To add a dual-coded overlay, call:

``nanviewer structural.nii.gz --overlay beta.nii.gz --overlay_alpha pval.nii.gz``

For discussion of other options, please see the :py:mod:`~nanslice.nanslicer`
documentation - the majority of options are the same across both programs.
"""
import sys
import argparse
import numpy as np
import scipy.ndimage.interpolation as ndinterp
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from PyQt5 import QtCore, QtWidgets
from .util import add_common_arguments
from .colorbar import colorbar, alphabar
from .slicer import Slicer, axis_indices, Axis_map
from .layer import Layer, blend_layers

PROG_NAME = 'NaNViewer'
PROG_VERSION = "1.0"

def crosshairs(axis, point, direction, orient, color='g'):
    """
    Helper function to draw crosshairs on an axis
    """
    ind1, ind2 = axis_indices(Axis_map[direction], orient)
    vline = axis.axvline(x=point[ind1], color=color)
    hline = axis.axhline(y=point[ind2], color=color)
    return (vline, hline)

def sample_point(img, point, order=1):
    """
    Helper function to sample an image at a single point (instead of a whole slice)
    """
    scale = np.mat(img.get_affine()[0:3, 0:3]).I
    offset = np.dot(-scale, img.get_affine()[0:3, 3]).T
    s_point = np.dot(scale, point).T + offset[:]
    return ndinterp.map_coordinates(img.get_data().squeeze(), s_point, order=order)


class NaNCanvas(FigureCanvas):
    """
    Canvas to draw slices in
    """

    def __init__(self, args, parent=None, width=5, height=4, dpi=100):
        self.layers = [Layer(args.base_image,
                             cmap=args.base_map,
                             clim=args.base_lims,
                             mask=args.mask,
                             interp_order=args.interp_order),]
        if args.overlay:
            self.layers.append(Layer(args.overlay,
                                     cmap=args.overlay_map,
                                     clim=args.overlay_lim,
                                     mask=args.overlay_mask,
                                     mask_threshold=args.overlay_mask_thresh,
                                     alpha=args.overlay_alpha,
                                     alpha_lim=args.overlay_alpha_lim,
                                     interp_order=args.interp_order))

        self.fig = Figure(figsize=(width, height), dpi=dpi, facecolor='k')
        FigureCanvas.__init__(self, self.fig)
        FigureCanvas.mpl_connect(self, 'button_press_event', self.handle_mouse_event)
        FigureCanvas.mpl_connect(self, 'motion_notify_event', self.handle_mouse_event)
        self.setParent(parent)
        FigureCanvas.setSizePolicy(self,
                                   QtWidgets.QSizePolicy.Expanding,
                                   QtWidgets.QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        gs1 = GridSpec(1, 3)
        if args.base_map or args.overlay:
            gs2 = GridSpec(1, 1)
            gs1.update(left=0.01, right=0.99, bottom=0.16, top=0.99, wspace=0.01, hspace=0.01)
            gs2.update(left=0.08, right=0.92, bottom=0.08, top=0.15, wspace=0.1, hspace=0.1)
            self.cbar_axis = self.fig.add_subplot(gs2[0], facecolor='black')
            if args.overlay_alpha:
                alphabar(self.cbar_axis, args.overlay_map, args.overlay_lim, args.overlay_label,
                         args.overlay_alpha_lim, args.overlay_alpha_label)
            else:
                if args.base_map:
                    colorbar(self.cbar_axis, self.layers[0].cmap,
                             self.layers[0].clim, args.base_label)
                else:
                    colorbar(self.cbar_axis, self.layers[1].cmap,
                             self.layers[1].clim, args.overlay_label)
        else:
            gs1.update(left=0.01, right=0.99, bottom=0.01, top=0.99, wspace=0.01, hspace=0.01)
        self.axes = [self.fig.add_subplot(gs, facecolor='black') for gs in gs1]
        self.cursor = self.layers[0].bbox.center
        self.args = args
        self._slices = [None, None, None]
        self._images = [None, None, None]
        self._contours = [None, None, None]
        self._crosshairs = [None, None, None]
        self._first_time = True
        self.directions = ('z', 'x', 'y')
        self.update_figure()

    def update_figure(self, hold=None):
        """
        Updates the three axis views
        """
        #t0 = time.time()
        # Save typing and lookup time
        args = self.args
        bbox = self.layers[0].bbox
        cursor = self.cursor
        directions = self.directions
        # Do these individually now because I'm not clever enough to set them in the loop
        if not self._first_time:
            for crosshair in self._crosshairs:
                # Crosshairs consist of a vline/hline pair
                crosshair[0].remove()
                crosshair[1].remove()
        for i in range(3):
            if i != hold:
                self._slices[i] = Slicer(bbox, cursor[i], directions[i],
                                         args.samples, orient=args.orient)
                sl_final = blend_layers(self.layers, self._slices[i])
                # Draw image
                if self._first_time:
                    self._images[i] = self.axes[i].imshow(sl_final, origin='lower',
                                                          extent=self._slices[i].extent,
                                                          interpolation=self.args.interp)
                    # If these calls go in __init__ then images don't show
                    self.axes[i].axis('off')
                    self.axes[i].axis('image')
                else:
                    self._images[i].set_data(sl_final)

                # Draw contours. For contours remove collection manually
                if self.args.contour:
                    if not self._first_time:
                        for coll in self._contours[i].collections:
                            coll.remove()
                    sl_contour = self.layers[1].get_slice(self._slices[i])
                    self._contours[i] = self.axes[i].contour(sl_contour, levels=self.args.contour,
                                                             colors=args.contour_color, linestyles=args.contour_style,
                                                             linewidths=1.0, origin='lower',
                                                             extent=self._slices[i].extent)
            self._crosshairs[i] = crosshairs(self.axes[i], self.cursor,
                                             directions[i], self.args.orient)
        self._first_time = False
        #print('Update time:', (time.time() - t0)*1000, 'ms')
        self.draw()

    def handle_mouse_event(self, event):
        """
        Updates the slice locations and crosshair
        """
        if event.button == 1:
            for i in range(3):
                if event.inaxes == self.axes[i]:
                    ind1, ind2 = axis_indices(Axis_map[self.directions[i]], self.args.orient)
                    self.cursor[ind1] = event.xdata
                    self.cursor[ind2] = event.ydata
                    self.update_figure(hold=i)
            msg = 'Cursor: ' + str(self.cursor)
            if len(self.layers) > 1:
                color_val = sample_point(self.layers[1].base_image,
                                         self.cursor,
                                         self.args.interp_order)
                msg = msg + ' ' + self.args.color_label + ': ' + str(color_val[0])
                if self.layers[1].alpha_image:
                    alpha_val = sample_point(self.layers[1].alpha_image,
                                             self.cursor,
                                             self.args.interp_order)
                    msg = msg + ' ' + self.args.alpha_label + ': ' + str(alpha_val[0])
            # Parent of this is the layout, call parent again to get the main window
            self.parent().parent().statusBar().showMessage(msg)

class NaNViewWindow(QtWidgets.QMainWindow):
    """
    Main window class for the viewer
    """
    def __init__(self, args):
        QtWidgets.QMainWindow.__init__(self)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setWindowTitle("NaNView")
        self.file_menu = QtWidgets.QMenu('&File', self)
        self.file_menu.addAction('&Quit', self._file_quit,
                                 QtCore.Qt.CTRL + QtCore.Qt.Key_Q)
        self.menuBar().addMenu(self.file_menu)
        self.help_menu = QtWidgets.QMenu('&Help', self)
        self.menuBar().addSeparator()
        self.menuBar().addMenu(self.help_menu)
        self.help_menu.addAction('&About', self._about)
        self.statusBar().showMessage("NaNViewer", 2000)
        self.main_widget = QtWidgets.QWidget(self)
        self.main_widget.setFocus()
        self.setCentralWidget(self.main_widget)
        layout = QtWidgets.QVBoxLayout(self.main_widget)
        layout.addWidget(NaNCanvas(args, self.main_widget, width=5, height=4, dpi=100))

    def _file_quit(self):
        """Callback for quit menu entry"""
        self.close()

    def _about(self):
        """Callback for about menu entry"""
        QtWidgets.QMessageBox.about(self, "About", """NaNViewer
Copyright 2017 Tobias Wood

A simple viewer for dual-coded overlays.

With thanks to http://matplotlib.org/examples/user_interfaces/embedding_in_qt5.html""")

def main(args=None):
    """
    Main function.

    Parameters:

    - args -- Command-line arguments. See module documentation or command-line for full list
    """
    parser = argparse.ArgumentParser(description='Dual-coding viewer')
    add_common_arguments(parser)
    args = parser.parse_args()
    application = QtWidgets.QApplication(sys.argv)
    window = NaNViewWindow(args)
    window.setWindowTitle("%s" % PROG_NAME)
    window.show()
    sys.exit(application.exec_())

if __name__ == "__main__":
    main()



"""Setup file to Not Another Neuroimaging Slicer

"""

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='nanslice',
    version='1.1',
    description='Scripts to slice and display neuroimages (probably stored in nifti format)',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/spinicist/nanslice',
    author='Tobias Wood',
    author_email='tobias@spinicist.org.uk',
    py_modules=['nanslice'],
    install_requires=['matplotlib>=3.2.0',
                      'nibabel>=2.0.0',
                      'numpy>=1.14.2',
                      'scipy>=1.1.0',
                      'colorcet>=2.0.0'],
    python_requires='>=3',
    license='MPL',
    classifiers=['Topic :: Scientific/Engineering :: Visualization',
                 'Programming Language :: Python :: 3',
                 ],
    keywords='neuroimaging nifti',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'nanslicer=nanslice.nanslicer:main',
            'nanviewer=nanslice.nanviewer:main',
            'nanscroll=nanslice.nanscroll:main'
        ],
    },
)



#!/usr/bin/env python
"""nanscroll.py

This implements a command-line utility (installed as nanscroll) which will create
a video scrolling through one axis of an image. This is installed by PIP as ``nanscroll``.

The majority of options are the same as :py:mod:`~nanslice.nanslicer`.
"""
import argparse
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.animation import FuncAnimation
from .box import Box
from .slicer import Slicer, Axis_map
from .layer import Layer, blend_layers
from .util import add_common_arguments


def main(args=None):
    """
    The main function for the utility.

    Parameters:

    - args -- The command line-arguments
    """
    parser = argparse.ArgumentParser(
        description="Makes a video scrolling through an image")
    add_common_arguments(parser)
    parser.add_argument('output', help='Output image name', type=str)
    parser.add_argument('--slices', type=int, default=-1,
                        help='Number of slices to scroll through')
    parser.add_argument('--slice_axis', type=str, default='z',
                        help='Axis to slice along (x/y/z)')
    parser.add_argument('--slice_lims', type=float, nargs=2, default=(0.01, 0.99),
                        help='Slice between these limits along the axis, default=0.1 0.9')
    parser.add_argument('--volume', type=int, default=0,
                        help='Use this volume from a timeseries')
    parser.add_argument('--figsize', type=float, nargs=2, default=(6, 6),
                        help='Figure size (width, height) in inches')
    parser.add_argument('--dpi', type=int, default=150,
                        help='DPI for output figure')
    parser.add_argument('--fps', type=int, default=8,
                        help='Framerate for video (default 8)')
    parser.add_argument('--bitrate', type=int,
                        default=2048, help='Encoder bit-rate')
    args = parser.parse_args()

    print('*** Loading files')
    print('Loading base image: ', args.base_image)
    layers = [Layer(args.base_image, cmap=args.base_map, clim=args.base_lims, mask=args.mask,
                    interp_order=args.interp_order, volume=args.volume), ]
    if args.overlay:
        layers.append(Layer(args.overlay, cmap=args.overlay_map, clim=args.overlay_lims,
                            mask=args.overlay_mask, mask_threshold=args.overlay_mask_thresh,
                            alpha=args.alpha, alpha_lims=args.alpha_lims,
                            interp_order=args.interp_order))

    print('*** Setup')
    if layers[0].mask_image:
        bbox = Box.fromMask(layers[0].mask_image)
    else:
        bbox = Box.fromImage(layers[0].image)
    print(bbox)
    args.slice_axis = Axis_map[args.slice_axis]
    if args.slices == -1:
        slices = layers[0].image.shape[args.slice_axis]
    else:
        slices = args.slices
    print(slices)
    slice_pos = bbox.start[args.slice_axis] + \
        bbox.diag[args.slice_axis] * np.linspace(args.slice_lims[0], args.slice_lims[1],
                                                 slices)
    if args.orient == 'preclin':
        origin = 'upper'
    else:
        origin = 'lower'

    gs1 = gridspec.GridSpec(1, 1)
    gs1.update(left=0.01, right=0.99, bottom=0.01,
               top=0.99, wspace=0.01, hspace=0.01)
    fig = plt.figure(facecolor='black', figsize=args.figsize, dpi=args.dpi)

    print('*** Init Frame')
    axes = plt.subplot(gs1[0], facecolor='black')
    slicer = Slicer(
        bbox, slice_pos[0], args.slice_axis, args.samples, orient=args.orient)
    sl_final = blend_layers(layers, slicer)
    image = axes.imshow(sl_final, origin=origin,
                        extent=slicer.extent, interpolation=args.interp)
    axes.axis('off')

    def update_frame(frame):
        """Draws the next frame"""
        print('Slice pos ', slice_pos[frame])
        slicer = Slicer(bbox, slice_pos[frame], args.slice_axis,
                        args.samples, orient=args.orient)
        sl_final = blend_layers(layers, slicer)
        image.set_data(sl_final)
    print('*** Animate Frame')
    ani = FuncAnimation(fig, update_frame, frames=len(slice_pos))
    print('*** Save')
    ani.save(args.output, fps=args.fps, bitrate=args.bitrate,
             savefig_kwargs={'facecolor': 'black'})


if __name__ == "__main__":
    main()
